import pysam
import time
import numpy as np

bam_path = "data/nsc_1_mat.bam"

with pysam.AlignmentFile(bam_path, "rb") as bam:
    start = time.time()
    for read in bam.fetch():
        data = read.query_name  # Minimal access
    print("Fetch qname:", time.time() - start, "seconds")

with pysam.AlignmentFile(bam_path, "rb") as bam:
    start = time.time()
    for read in bam.fetch():
        data = read.query_sequence
    print("Fetch qseq:", time.time() - start, "seconds")

def qual_to_fastq(qualities):
    """Convert query_qualities to a FASTQ QUAL string using bytearray (FAST)."""
    return bytearray(q + 33 for q in qualities).decode()

def qual_to_fastq_numpy(qualities):
    """Convert query_qualities to FASTQ QUAL using NumPy (Best for Large Arrays)."""
    return (np.array(qualities, dtype=np.uint8) + 33).tobytes().decode()

with pysam.AlignmentFile(bam_path, "rb") as bam:
    start = time.time()
    for read in bam.fetch():
        data = qual_to_fastq_numpy(read.query_qualities)
    print("Fetch qual:", time.time() - start, "seconds")



import pysam
import numpy as np
from dataclasses import dataclass

@dataclass
class Read:
    """Class to represent a FASTQ read."""
    read_id: str
    sequence: str
    quality: str

    def __init__(self, read_id, sequence, quality):
        self.read_id = read_id
        self.sequence = sequence
        self.quality = quality
        self.validate()

    def __len__(self):
        return len(self.sequence)
    
    # validate the read
    def validate(self):
        if len(self.sequence) != len(self.quality):
            raise ValueError("Sequence and quality strings must be of equal length.")
        
    def is_valid(self):
        try:
            self.validate()
            return True
        except ValueError:
            return False
        
    def to_fastq(self):
        return f"@{self.read_id}\n{self.sequence}\n+\n{self.quality}"

def count_reads(bam_file):
    """Count the number of reads in a BAM file."""
    with pysam.AlignmentFile(bam_file, "rb") as bam:
        return sum(1 for _ in bam)

def qual_to_fastq_numpy(qualities):
    """Convert query_qualities to FASTQ QUAL using NumPy (Best for Large Arrays)."""
    return (np.array(qualities, dtype=np.uint8) + 33).tobytes().decode()

def parse_reads(bam_file):
    """Parse reads from a BAM file."""
    with pysam.AlignmentFile(bam_file, "rb") as bam:
        for read in bam:
            yield Read(read.query_name, read.query_sequence, qual_to_fastq_numpy(read.query_qualities))

def keep_n_bases(read, n, on = "left"):
    """Trim n bases of a read."""
    read_len = len(read)

    if n > read_len:
        return read
    elif on == "left":
        seq = read.sequence[:n]
        qual = read.quality[:n]
    elif on == "right":
        seq = read.sequence[(read_len - n):]
        qual = read.quality[(read_len - n):]
    
    return Read(read.read_id, seq, qual)
    
def split_read(read, at: list[int]):
    """Split a read at a given positions."""
    reads = []

    if len(at) == 0:
        read.read_id = f'{read.read_id}_0'
        return [read]
    
    count = 0
    start = 0

    for pos in at:
        reads.append(Read(f'{read.read_id}_{count}', read.sequence[start:pos], read.quality[start:pos]))
        start = pos
        count += 1
    
    reads.append(Read(f'{read.read_id}_{count}', read.sequence[start:], read.quality[start:]))

    return reads

def calculate_split(read, target_len: int):
    """Calculate split locations for a read given a target length"""
    if len(read) < target_len:
        return []

    # find the number of splits
    split_loc = round(len(read) / target_len)
    split_size = len(read) // split_loc

    return [i * split_size for i in range(1, split_loc)]




import argparse
import sys
import time
from bamurai.core import *

def split_reads(args):
    # record runtime
    print("Running Bamurai split...", file=sys.stderr)
    start_time = time.time()

    # pretty print the arguments in arg: value format
    arg_desc_dict = {
        "reads": "Input file",
        "len_target": "Target length",
        "output": "Output file"
    }
    print("Arguments:", file=sys.stderr)
    for arg, value in vars(args).items():
        if arg in arg_desc_dict:
            print(f"  {arg_desc_dict[arg]}: {value}", file=sys.stderr)

    # Read the input reads file
    read_lens = []

    # clear the output file
    if args.output:
        f = open(args.output, "w")

    for read in parse_reads(args.reads):
        split_locs = calculate_split(read, target_len = args.len_target)
        split = split_read(read, at = split_locs)
        for read in split:
            read_lens.append(len(read))
            
            if args.output:
                f.write(read.to_fastq())
                f.write("\n")
            else:
                print(read.to_fastq())

    if args.output:
        f.close()

    avg_read_len = round(sum(read_lens) / len(read_lens))
    print(f"Average split read length: {avg_read_len}", file=sys.stderr)
    print(f"Time taken: {round(time.time() - start_time, 2)} seconds", file=sys.stderr)

def main():
    parser = argparse.ArgumentParser(description="Bamurai: A tool for processing BAM files")
    subparsers = parser.add_subparsers(dest="command")

    # Subparser for the "split" command
    parser_split = subparsers.add_parser("split", help="Split reads in a BAM file to a target length")
    parser_split.add_argument("reads", type=str, help="Input reads file (BAM)")
    parser_split.add_argument("-l", "--len_target", type=int, help="Target length for splitting reads")
    parser_split.add_argument("-o", "--output", type=str, nargs='?', help="Output file")
    parser_split.set_defaults(func=split_reads)

    args = parser.parse_args()
    if args.command:
        args.func(args)
    else:
        parser.print_help()

if __name__ == "__main__":
    main()


from setuptools import setup, find_packages

setup(
    name="bamurai",
    version="0.1.0",
    packages=find_packages(),
    include_package_data=False,
    entry_points={
        "console_scripts": [
            "bamurai=bamurai.cli:main",
        ],
    },
    license_files=('LICENSE',),
)